package ru.dbotthepony.kstarbound.client.gl

import org.lwjgl.opengl.GL46.*
import org.lwjgl.system.MemoryUtil
import ru.dbotthepony.kstarbound.client.StarboundClient
import java.nio.ByteBuffer

sealed class BufferObject(val glType: Int) : GLObject() {
	final override val client = StarboundClient.current()
	final override val pointer = glGenBuffers()

	init {
		checkForGLError("Creating Vertex Buffer Object")
		client.registerCleanable(this, ::glDeleteBuffers, pointer)
	}

	fun bufferData(data: ByteBuffer, usage: Int): BufferObject {
		client.ensureSameThread()
		glNamedBufferData(pointer, data, usage)
		checkForGLError()
		return this
	}

	fun bufferData(data: ByteBuffer, usage: Int, length: Long): BufferObject {
		client.ensureSameThread()

		if (length > data.remaining().toLong()) {
			throw IndexOutOfBoundsException("Tried to upload $data into $pointer with offset at ${data.position()} with length of $length, but that is longer than remaining data length of ${data.remaining()}!")
		}

		nglNamedBufferData(pointer, length, MemoryUtil.memAddress(data), usage)

		checkForGLError()
		return this
	}

	fun bufferData(data: IntArray, usage: Int): BufferObject {
		client.ensureSameThread()
		glNamedBufferData(pointer, data, usage)
		checkForGLError()
		return this
	}

	fun bufferData(data: FloatArray, usage: Int): BufferObject {
		client.ensureSameThread()
		glNamedBufferData(pointer, data, usage)
		checkForGLError()
		return this
	}

	fun bufferData(data: DoubleArray, usage: Int): BufferObject {
		client.ensureSameThread()
		glNamedBufferData(pointer, data, usage)
		checkForGLError()
		return this
	}

	fun bufferData(data: LongArray, usage: Int): BufferObject {
		client.ensureSameThread()
		glNamedBufferData(pointer, data, usage)
		checkForGLError()
		return this
	}

	class VBO : BufferObject(GL_ARRAY_BUFFER) {
		override fun bind() {
			client.vbo = this
		}

		override fun unbind() {
			if (client.vbo == this) client.vbo = null
		}
	}

	class EBO : BufferObject(GL_ELEMENT_ARRAY_BUFFER) {
		override fun bind() {
			client.ebo = this
		}

		override fun unbind() {
			if (client.ebo == this) client.ebo = null
		}
	}
}
